# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from math import atan2, degrees

from mathmakerlib import required, config
from mathmakerlib.LaTeX import MATHEMATICAL_NOTATIONS
from mathmakerlib.exceptions import ZeroVector
from mathmakerlib.core.oriented import Oriented
from mathmakerlib.core.oriented import check_winding, shoelace_formula
from mathmakerlib.core.drawable import Colored, HasThickness, HasRadius
from mathmakerlib.core.drawable import Labeled, HasArrowTips
from mathmakerlib.core.drawable import tikz_options_list, Drawable
from mathmakerlib.core.drawable import tikz_approx_position
from mathmakerlib.core.dimensional import Dimensional
from mathmakerlib.geometry.point import Point
from mathmakerlib.geometry.vector import Vector
from mathmakerlib.geometry.bipoint import Bipoint
from mathmakerlib.calculus.number import Number, is_number

AVAILABLE_NAMING_MODES = ['from_endpoints', 'from_armspoints', 'from_vertex']


class AngleDecoration(Labeled, Colored, HasThickness, HasRadius, HasArrowTips):

    def __init__(self, color=None, thickness='thick', label='default',
                 radius=Number('0.25', unit='cm'), variety='single',
                 gap=Number('0.4', unit='cm'), eccentricity='automatic',
                 hatchmark=None, do_draw=True, arrow_tips=None):
        self.do_draw = do_draw
        self.arrow_tips = arrow_tips
        self.color = color
        self.thickness = thickness
        self.label = label
        self.radius = radius
        self.gap = gap
        if gap is None:
            self.gap = None
        else:
            u = self.radius.unit if isinstance(self.radius, Number) else None
            self.gap = Number(gap, unit=u)
        # Eccentricity must be set *after* radius, in order to be able to
        # calculate a reasonable default eccentricity based on the radius
        self.eccentricity = eccentricity
        self.variety = variety
        self.hatchmark = hatchmark

    def __repr__(self):
        return 'AngleDecoration(variety={}; hatchmark={}; label={}; '\
            'color={}; thickness={}; radius={}; eccentricity={})'\
            .format(self.variety, self.hatchmark, self.label, self.color,
                    self.thickness, str(self.radius), self.eccentricity)

    @property
    def do_draw(self):
        return self._do_draw

    @do_draw.setter
    def do_draw(self, value):
        if isinstance(value, bool):
            self._do_draw = value
        else:
            raise TypeError('do_draw must be a boolean; '
                            'got {} instead.'.format(type(value)))

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, value):
        if is_number(value):
            self._radius = Number(value)
        elif value is None:
            self._radius = None
        else:
            raise TypeError('Expected a number as radius. Got {} instead.'
                            .format(str(type(value))))
        if hasattr(self, '_eccentricity') and hasattr(self, '_gap'):
            self.eccentricity = 'automatic'

    @property
    def eccentricity(self):
        return self._eccentricity

    @eccentricity.setter
    def eccentricity(self, value):
        if value == 'automatic':
            if self.gap is None:
                raise ValueError('Cannot calculate the eccentricity if gap '
                                 'is None.')
            value = (self.gap / self.radius + 1)\
                .rounded(Number('0.01')).standardized()
        if not (value is None or is_number(value)):
            raise TypeError('The eccentricity of an AngleDecoration must be '
                            'None or a Number. Found {} instead.'
                            .format(type(value)))
        self._eccentricity = value

    @property
    def variety(self):
        return self._variety

    @variety.setter
    def variety(self, value):
        if value not in [None, 'single', 'double', 'triple']:
            raise TypeError('AngleDecoration\'s variety can be None, '
                            '\'single\', \'double\' or \'triple\'. '
                            'Found {} instead (type: {}).'
                            .format(repr(value), type(value)))
        self._variety = value

    @property
    def gap(self):
        return self._gap

    @gap.setter
    def gap(self, value):
        if not (is_number(value) or value is None):
            raise TypeError('The gap value must be None or a number. '
                            'Found {} instead (type: {}).'
                            .format(repr(value), type(value)))
        if value is None:
            self._gap = None
        else:
            self._gap = Number(value)

    @property
    def hatchmark(self):
        return self._hatchmark

    @hatchmark.setter
    def hatchmark(self, value):
        if value not in [None, 'singledash', 'doubledash', 'tripledash']:
            raise TypeError('AngleDecoration\'s hatchmark can be None, '
                            '\'singledash\', \'doubledash\' or '
                            '\'tripledash\'. Found {} instead (type: {}).'
                            .format(repr(value), type(value)))
        self._hatchmark = value

    def tikz_attributes(self, radius_coeff=1, do_label=True):
        if not is_number(radius_coeff):
            raise TypeError('radius_coeff must be a number, found {} instead.'
                            .format(type(radius_coeff)))
        attributes = []
        if do_label and self.label not in [None, 'default']:
            required.tikz_library['quotes'] = True
            attributes.append('"{}"'.format(self.label))
            if self.eccentricity is not None:
                attributes.append('angle eccentricity={}'
                                  .format(self.eccentricity))
        if self.variety is not None:
            if self.do_draw:
                attributes.append('draw')
            if self.arrow_tips is not None:
                attributes.append(self.arrow_tips)
            if self.thickness is not None:
                attributes.append(self.thickness)
            if self.radius is not None:
                attributes.append('angle radius = {}'
                                  .format((self.radius * radius_coeff)
                                          .rounded(Number('0.01')).uiprinted))
            if self.hatchmark is not None:
                attributes.append(self.hatchmark)
                required.tikz_library['decorations.markings'] = True
                required.tikzset[self.hatchmark + '_hatchmark'] = True
        if (self.variety is not None
            or (do_label and self.label not in [None, 'default'])):
            if self.color is not None:
                attributes.append(self.color)
        return '[{}]'.format(', '.join(attributes))

    def generate_tikz(self, *points_names):
        if not len(points_names) == 3:
            raise RuntimeError('Three Points\' names must be provided to '
                               'generate the AngleDecoration. Found {} '
                               'arguments instead.'.format(len(points_names)))
        last_layer = {None: 1, 'single': 1, 'double': 2,
                      'triple': 3}[self.variety]
        pic_attr = self.tikz_attributes(do_label=last_layer == 1)
        if pic_attr == '[]':
            return ''
        required.tikz_library['angles'] = True
        deco = ['pic {} {{angle = {}--{}--{}}}'
                .format(pic_attr, *points_names)]
        if self.variety in ['double', 'triple']:
            space_sep = Number('0.16')
            deco.append('pic {} {{angle = {}--{}--{}}}'
                        .format(self.tikz_attributes(
                                radius_coeff=1 + space_sep,
                                do_label=last_layer == 2),
                                *points_names))
            if self.variety == 'triple':
                deco.append('pic {} {{angle = {}--{}--{}}}'
                            .format(self.tikz_attributes(
                                    radius_coeff=1 + 2 * space_sep,
                                    do_label=last_layer == 3),
                                    *points_names))
        return deco


class Angle(Drawable, Oriented, HasThickness, Dimensional):

    def __init__(self, point, vertex, point_or_measure, decoration=None,
                 mark_right=False, second_point_name='auto', label=None,
                 color=None, thickness='thick', armspoints=None,
                 label_vertex=False, draw_vertex=False,
                 label_armspoints=False, draw_armspoints=False,
                 label_endpoints=False, draw_endpoints=False,
                 naming_mode='from_endpoints', decoration2=None):
        """
        :param point: a Point of an arm of the Angle
        :type point: Point
        :param vertex: the Angle's vertex
        :type vertex: Point
        :param point_or_measure: either a Point of the other arm of the Angle,
        or the measure of the Angle
        :type point_or_measure: Point or number
        :param decoration: the decoration of the Angle
        :type decoration: None or AngleDecoration
        :param mark_right: to tell whether to mark the angle as a right angle
        :type mark_right: bool
        :param second_point_name: Only used if point_or_measure is a measure,
        this is the name of the 2d arm's Point. If set to 'auto', then the name
        of the first Point will be used, concatenated to a '.
        :type second_point_name: str
        :param thickness: the Angle's arms' thickness. Available values are
        TikZ's ones.
        :type thickness: str
        :param color: the color of the Angle's arms.
        :type color: str
        :param naming_mode: how to build the name. Possible modes are:
        'from_endpoints', 'from_armspoints', 'from_vertex'. Note that if no
        armspoints are defined, then trying to get the Angle.name will raise an
        error
        :type naming_mode: str
        """
        self.color = color
        self.thickness = thickness
        self.naming_mode = naming_mode
        self.decoration = decoration
        self.decoration2 = decoration2
        # The label must be set *after* the possible decoration, because it
        # will actually be handled by self.decoration
        if (self.decoration is None
            or self.decoration.label in [None, 'default']):
            self.label = label
        else:
            if label is not None:
                raise ValueError('The label has been set twice, as Angle\'s '
                                 'keyword argument ({}) and as its '
                                 'AngleDecoration\'s keyword argument ({}).'
                                 .format(repr(label),
                                         repr(self.decoration.label_value)))
        self.mark_right = mark_right
        self.label_vertex = label_vertex
        self.label_endpoints = label_endpoints
        self.draw_endpoints = draw_endpoints
        self.label_armspoints = label_armspoints
        self.draw_armspoints = draw_armspoints
        self.draw_vertex = draw_vertex
        if not (isinstance(point, Point)
                and isinstance(vertex, Point)
                and (isinstance(point_or_measure, Point)
                     or is_number(point_or_measure))):
            raise TypeError('Three Points, or two Points and the measure of '
                            'the angle are required to build an Angle. '
                            'Found instead: {}, {} and {}.'
                            .format(type(point), type(vertex),
                                    type(point_or_measure)))
        self._points = [point, vertex]
        if isinstance(point_or_measure, Point):
            self._points.append(point_or_measure)
        else:
            self._points.append(point.rotate(vertex, point_or_measure,
                                             rename=second_point_name))

        if any([p.three_dimensional for p in self._points]):
            self._three_dimensional = True
        else:
            self._three_dimensional = False

        # Measure of the angle:
        if self._three_dimensional:
            u = Vector(self.points[1], self.points[0])
            v = Vector(self.points[1], self.points[2])
            self._measure = Number(str(degrees(atan2(u.cross(v).length,
                                                     u.dot(v)))))
        else:  # 2D angles measure
            p0 = Point(self._points[0].x - self._points[1].x,
                       self._points[0].y - self._points[1].y,
                       None)
            p2 = Point(self._points[2].x - self._points[1].x,
                       self._points[2].y - self._points[1].y,
                       None)
            α0 = Number(str(degrees(atan2(p0.y, p0.x))))
            α2 = Number(str(degrees(atan2(p2.y, p2.x))))
            self._measure = α2 - α0

        if self._measure < 0:
            self._measure += 360

        # This is not like the matching Triangle!
        if shoelace_formula(*self.points) > 0:
            self.winding = 'clockwise'
        else:
            self.winding = 'anticlockwise'

        arm0 = Bipoint(self._points[1], self._points[0])
        arm1 = Bipoint(self._points[1], self._points[2])
        self._arms = [arm0, arm1]
        self.armspoints = armspoints

        # Only 2D: labels positioning
        if not self.three_dimensional:
            # Vertex' label positioning
            bisector = Vector(self._points[0], self.vertex)\
                .bisector(Vector(self._points[2], self.vertex),
                          new_endpoint_name=None)
            try:
                self._points[1].label_position = \
                    tikz_approx_position(bisector.slope360)
            except ZeroVector:
                self._points[1].label_position = \
                    tikz_approx_position(
                        Bipoint(self.vertex,
                                self._points[0].rotate(self.vertex, -90,
                                                       rename=None)
                                ).slope360)

            # Endpoints labels positioning
            direction = 1 if self.winding == 'anticlockwise' else -1
            self.endpoints[0].label_position = \
                tikz_approx_position(arm0.slope360 - direction * 55)
            self.endpoints[1].label_position = \
                tikz_approx_position(arm1.slope360 + direction * 55)

    def __repr__(self):
        return 'Angle({}, {}, {})'\
            .format(self.points[0].name, self.points[1].name,
                    self.points[2].name)

    @property
    def vertex(self):
        return self._points[1]

    @property
    def points(self):
        return self._points

    @property
    def endpoints(self):
        return [self._points[0], self._points[2]]

    @property
    def armspoints(self):
        return self._armspoints

    @armspoints.setter
    def armspoints(self, value):
        self._armspoints = []
        if value is None:
            value = []
        if not isinstance(value, list):
            raise TypeError('A list must be provided to setup armspoints. '
                            'Found {} instead.'.format(type(value)))
        if len(value) > len(self.arms):
            raise ValueError('More values are provided ({}) then available '
                             'arms ({}).'.format(len(value), len(self.arms)))
        for i, p in enumerate(value):
            if not isinstance(p, tuple):
                raise TypeError('Each arm\'s point must be defined by a '
                                'tuple. Found {} instead.'.format(type(p)))
            if len(p) > 2:
                raise TypeError('Each arm\'s point must be defined by a '
                                'tuple of 1 or 2 elements. '
                                'Found {} elements instead.'.format(len(p)))
            if p[0] in [None, '']:
                name = 'automatic'
            else:
                name = p[0]
            try:
                position = p[1]
            except IndexError:
                position = config.angles.DEFAULT_ARMSPOINTS_POSITION
            self._armspoints.append(self.arms[i].point_at(position, name))
        if len(self._armspoints):
            self.label_armspoints = True
            self.draw_armspoints = True

        # Armspoints labels positioning
        direction = 1 if self.winding == 'anticlockwise' else -1
        for i, ap in enumerate(self.armspoints):
            xdir = -1 if i == 0 else 1
            self.armspoints[i].label_position = \
                tikz_approx_position(self.arms[i].slope360
                                     + xdir * direction * 60)

    @property
    def arms(self):
        return self._arms

    @property
    def measure(self):
        """Measure of the Angle."""
        return self._measure

    @property
    def label(self):
        """
        The label of an Angle is handled internally by its AngleDecoration.
        """
        if hasattr(self, '_decoration') and self.decoration is not None:
            return self.decoration.label
        else:
            return None

    @label.setter
    def label(self, value):
        """
        The label of an Angle is handled internally by its AngleDecoration.
        """
        if self.decoration is None:
            self.decoration = AngleDecoration(variety=None, label=value)
        else:
            self.decoration.label = value

    @property
    def decoration(self):
        return self._decoration

    @decoration.setter
    def decoration(self, deco):
        if not (deco is None or isinstance(deco, AngleDecoration)):
            raise TypeError('An angle decoration must be None or belong to '
                            'the AngleDecoration class. Got {} instead.'
                            .format(type(deco)))
        if deco is None:
            if not hasattr(self, '_decoration') or self.label is None:
                self._decoration = None
            else:
                self._decoration = AngleDecoration(variety=None,
                                                   label=self.label)
        else:  # deco is an AngleDecoration
            # If the label has been set prior to an additional decoration,
            # keep it:
            if deco.label == 'default':
                deco.label = self.label
            self._decoration = deco

    @property
    def decoration2(self):
        return self._decoration2

    @decoration2.setter
    def decoration2(self, deco):
        if not (deco is None or isinstance(deco, AngleDecoration)):
            raise TypeError('An angle decoration must be None or belong to '
                            'the AngleDecoration class. Got {} instead.'
                            .format(type(deco)))
        self._decoration2 = deco

    @property
    def mark_right(self):
        return self._mark_right

    @mark_right.setter
    def mark_right(self, value):
        if not isinstance(value, bool):
            raise TypeError('\'mark_right\' must be a boolean')
        self._mark_right = value

    @property
    def draw_vertex(self):
        return self._draw_vertex

    @draw_vertex.setter
    def draw_vertex(self, value):
        if isinstance(value, bool):
            self._draw_vertex = value
        else:
            raise TypeError('draw_vertex must be a boolean; '
                            'got {} instead.'.format(type(value)))

    @property
    def draw_endpoints(self):
        return self._draw_endpoints

    @draw_endpoints.setter
    def draw_endpoints(self, value):
        if isinstance(value, bool):
            self._draw_endpoints = value
        else:
            raise TypeError('draw_endpoints must be a boolean; '
                            'got {} instead.'.format(type(value)))

    @property
    def draw_armspoints(self):
        return self._draw_armspoints

    @draw_armspoints.setter
    def draw_armspoints(self, value):
        if isinstance(value, bool):
            self._draw_armspoints = value
        else:
            raise TypeError('draw_armspoints must be a boolean; '
                            'got {} instead.'.format(type(value)))

    @property
    def label_vertex(self):
        return self._label_vertex

    @label_vertex.setter
    def label_vertex(self, value):
        if isinstance(value, bool):
            self._label_vertex = value
        else:
            raise TypeError('label_vertex must be a boolean; '
                            'got {} instead.'.format(type(value)))

    @property
    def label_endpoints(self):
        return self._label_endpoints

    @label_endpoints.setter
    def label_endpoints(self, value):
        if isinstance(value, bool):
            self._label_endpoints = value
        else:
            raise TypeError('label_endpoints must be a boolean; '
                            'got {} instead.'.format(type(value)))

    @property
    def label_armspoints(self):
        return self._label_armspoints

    @label_armspoints.setter
    def label_armspoints(self, value):
        if isinstance(value, bool):
            self._label_armspoints = value
        else:
            raise TypeError('label_armspoints must be a boolean; '
                            'got {} instead.'.format(type(value)))

    @property
    def naming_mode(self):
        return self._naming_mode

    @naming_mode.setter
    def naming_mode(self, value):
        if value not in AVAILABLE_NAMING_MODES:
            raise ValueError('naming_mode must belong to {}. Found {} instead.'
                             .format(AVAILABLE_NAMING_MODES, repr(value)))
        self._naming_mode = value

    @property
    def name(self):
        loc = None
        for lg in ['fr', 'en']:
            if config.language.startswith(lg):
                loc = lg
        if self.naming_mode == 'from_endpoints':
            content = '{}{}{}'.format(self.endpoints[0].name,
                                      self.vertex.name,
                                      self.endpoints[1].name)
        elif self.naming_mode == 'from_armspoints':
            if not self.armspoints:
                raise RuntimeError('The naming mode of this Angle is '
                                   '\'from_armspoints\' but the armspoints '
                                   'are not defined (empty list).')
            content = '{}{}{}'.format(self.armspoints[0].name,
                                      self.vertex.name,
                                      self.armspoints[1].name)
        elif self.naming_mode == 'from_vertex':
            content = self.vertex.name
        _name = MATHEMATICAL_NOTATIONS[loc]['angle_name']\
            .format(content=content)
        if 'stackon' in _name:
            required.package['stackengine'] = True
        if 'hstretch' in _name or 'vstretch' in _name:
            required.package['scalerel'] = True
        return _name

    def tikz_decorations(self):
        output_elements = []
        if not(self.decoration is None
               or (self.mark_right and self.label is None)):
            output_elements = \
                self.decoration.generate_tikz(self.points[0].name,
                                              self.vertex.name,
                                              self.points[2].name)
        if self.decoration2 is not None:
            output_elements += \
                self.decoration2.generate_tikz(self.points[0].name,
                                               self.vertex.name,
                                               self.points[2].name)
        return '\n'.join(output_elements)

    def tikz_rightangle_mark(self, winding='anticlockwise'):
        if self.decoration is None or not self.mark_right:
            return ''
        check_winding(winding)
        # Decimal numbers in TikZ must be written with a dot as decimal point.
        # As of now, there is no reliable way to temporarily change the
        # locale to 'C' (or 'en_US'), so here's a little patch that will
        # replace possibly other decimal points by a '.'.
        theta = Bipoint(self.vertex, self.points[0])\
            .slope.rounded(Number('0.01')).imprint(dot=True)
        rt = 'cm={{cos({θ}), sin({θ}), -sin({θ}), cos({θ}), ({v})}}' \
            .format(θ=theta, v=self.vertex.name)
        draw_options = tikz_options_list([self.decoration.thickness,
                                          self.decoration.color,
                                          rt])
        if winding == 'anticlockwise':
            rightangle_shape = '({R}, 0) -- ({R}, {R}) -- (0, {R})'\
                .format(R=self.decoration.radius.uiprinted)
        elif winding == 'clockwise':
            rightangle_shape = '({R}, 0) -- ({R}, -{R}) -- (0, -{R})'\
                .format(R=self.decoration.radius.uiprinted)
        return r'\draw' + '{} {};'.format(draw_options, rightangle_shape)

    def tikz_declarations(self):
        """Return the Points declarations."""
        points = self.points
        if len(self.armspoints):
            points = points + self.armspoints
        return '\n'.join([v.tikz_declarations() for v in points])

    def _tikz_draw_options(self):
        """
        The list of possible options for draw command.

        :rtype: list
        """
        return [self.thickness, self.color]

    def _tikz_draw_vertex(self):
        return self.vertex.tikz_draw()[0]

    def _tikz_draw_armspoints(self, nb=None):
        if nb is None:
            nb = [n for n in range(len(self.armspoints))]
        return '\n'.join([self.armspoints[n].tikz_draw()[0] for n in nb])

    def _tikz_draw_endpoints(self, nb=None):
        if nb is None:
            nb = [0, 1]
        return '\n'.join([self.endpoints[n].tikz_draw()[0] for n in nb])

    def tikz_drawing_comment(self):
        """
        Return the comments matching each drawing category.

        :rtype: list
        """
        comments = ['% Draw Angle']
        if self.draw_vertex:
            comments.append('% Draw Vertex')
        if self.draw_armspoints and len(self.armspoints):
            comments.append('% Draw Arms\' Points')
        if self.draw_endpoints:
            comments.append('% Draw End Points')
        if self.mark_right:
            comments.append('\n% Mark right angle')
        return comments

    def tikz_draw(self):
        """
        Return the commands to actually draw the object.

        If enabled, the vertex is drawn;
        if enabled the end Points are drawn;
        if enabled the arms' Points are drawn.
        The Angle is always drawn.

        :rtype: list
        """
        decoration = self.tikz_decorations()
        if decoration != '':
            decoration = '\n' + decoration
        commands = [r'\draw' + '{} ({}) -- ({}) -- ({}){};'
                    .format(tikz_options_list('draw', self),
                            *[p.name for p in self.points],
                            decoration)
                    ]
        if self.draw_vertex:
            commands.append(self._tikz_draw_vertex())
        if self.draw_armspoints and len(self.armspoints):
            commands.append(self._tikz_draw_armspoints())
        if self.draw_endpoints:
            commands.append(self._tikz_draw_endpoints())
        if self.mark_right:
            commands.append(self.tikz_rightangle_mark())
        return commands

    def tikz_label(self):
        """The Angle's label is included in its draw command."""

    def tikz_points_labels(self):
        """Return the command to write the object's points' labels."""
        labels = []
        if self.label_vertex:
            labels.append(self.vertex.tikz_label())
        if self.label_endpoints:
            labels.append(self.endpoints[0].tikz_label())
            labels.append(self.endpoints[1].tikz_label())
        if self.label_armspoints:
            labels.append(self.armspoints[0].tikz_label())
            labels.append(self.armspoints[1].tikz_label())
        return '\n'.join(labels)


class AnglesSet(Drawable):

    def __init__(self, *angles):
        self.angles = angles

    @property
    def angles(self):
        return self._angles

    @angles.setter
    def angles(self, angles):
        for α in angles:
            if not isinstance(α, Angle):
                raise TypeError('Any element of an AnglesSet must be an '
                                'Angle. Found {} instead.'.format(type(α)))
        self._angles = angles

    def tikz_declarations(self):
        """
        Return the necessary declarations (e.g. Points declarations).

        :rtype: str
        """
        points = []
        for α in self.angles:
            if α.vertex not in points:
                points.append(α.vertex)
            for p in α.endpoints:
                if p not in points:
                    points.append(p)
            for p in α.armspoints:
                if p not in points:
                    if α.draw_armspoints or α.label_armspoints:
                        points.append(p)
        points_names = [p.name for p in points]
        if len(points_names) != len(set(points_names)):
            raise RuntimeError('Two different Points have been provided the '
                               'same name in this list: {}'
                               .format('; '.join([str(p) for p in points])))
        return '\n'.join([p.tikz_declarations() for p in points])

    def _tikz_draw_options(self):
        """
        Each Angle will have its own options, so nothing to return.

        :rtype: list
        """
        return []

    def tikz_drawing_comment(self):
        """
        Return the comments matching each drawing category.

        :rtype: list
        """
        comments = ['% Draw Angles']
        vertices_to_draw = len([α.draw_vertex
                                for α in self.angles if α.draw_vertex])
        if vertices_to_draw:
            if vertices_to_draw == 1:
                comments.append('% Draw Vertex')
            else:
                comments.append('% Draw Vertices')
        if any([α.draw_armspoints and len(α.armspoints)
                for α in self.angles]):
            comments.append('% Draw Arms\' Points')
        if any([α.draw_endpoints for α in self.angles]):
            comments.append('% Draw End Points')
        if any([θ.mark_right for θ in self.angles]):
            comments.append('% Mark right Angles')
        return comments

    def tikz_label(self):
        """All labels are included in the Angles' draw commands."""

    def tikz_points_labels(self):
        """Return the command to write the Angles' points' labels."""
        labels = []
        labeled_points_names = []
        for α in self.angles:
            if α.label_vertex and α.vertex.name not in labeled_points_names:
                labels.append(α.vertex.tikz_label())
                labeled_points_names.append(α.vertex.name)
            if α.label_endpoints:
                for p in α.endpoints:
                    if p.name not in labeled_points_names:
                        labels.append(p.tikz_label())
                        labeled_points_names.append(p.name)
            if α.label_armspoints:
                for p in α.armspoints:
                    if p.name not in labeled_points_names:
                        labels.append(p.tikz_label())
                        labeled_points_names.append(p.name)
        return '\n'.join(labels)

    def _tikz_draw_right_angles_marks(self):
        return '\n'.join([θ.tikz_rightangle_mark(θ.winding)
                          for θ in self.angles
                          if θ.tikz_rightangle_mark(θ.winding) != ''])

    def tikz_draw(self):
        """
        Return the commands to actually draw the Angles.

        :rtype: list
        """
        angles_cmd = []
        for α in self.angles:
            decoration = α.tikz_decorations()
            if decoration != '':
                decoration = '\n' + decoration
            angles_cmd.append(r'\draw' + '{} ({}) -- ({}) -- ({}){};'
                              .format(tikz_options_list('draw', α),
                                      *[p.name for p in α.points],
                                      decoration))
        commands = ['\n'.join(angles_cmd)]

        labeled_points_names = []
        vertices_cmd = []
        for α in self.angles:
            if α.draw_vertex and α.vertex.name not in labeled_points_names:
                vertices_cmd.append(α._tikz_draw_vertex())
                labeled_points_names.append(α.vertex.name)
        vertices_cmd = '\n'.join(vertices_cmd)
        if vertices_cmd != '':
            commands.append(vertices_cmd)

        armspoints_cmd = []
        for α in self.angles:
            if α.draw_armspoints and len(α.armspoints):
                nb = []
                for i, p in enumerate(α.armspoints):
                    if p.name not in labeled_points_names:
                        nb.append(i)
                        labeled_points_names.append(p.name)
                armspoints_cmd.append(α._tikz_draw_armspoints(nb=nb))
        armspoints_cmd = '\n'.join(armspoints_cmd)
        if armspoints_cmd != '':
            commands.append(armspoints_cmd)

        endpoints_cmd = []
        for α in self.angles:
            if α.draw_endpoints and len(α.endpoints):
                nb = []
                for i, p in enumerate(α.endpoints):
                    if p.name not in labeled_points_names:
                        nb.append(i)
                        labeled_points_names.append(p.name)
                endpoints_cmd.append(α._tikz_draw_endpoints(nb=nb))
        endpoints_cmd = '\n'.join(endpoints_cmd)
        if endpoints_cmd != '':
            commands.append(endpoints_cmd)

        if any([θ.mark_right for θ in self.angles]):
            commands.append(self._tikz_draw_right_angles_marks())

        return commands



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from mathmakerlib.calculus.number import Number
from . import Rectangle, Rhombus, Equilateral


class Square(Rectangle, Rhombus):
    """Squares."""

    def __init__(self, start_vertex=None, name=None,
                 side_length=Number('1'),
                 mark_right_angles=True,
                 mark_equal_sides=True, use_mark='||',
                 draw_vertices=False, label_vertices=True,
                 thickness='thick', color=None, rotation_angle=0,
                 winding=None, sloped_sides_labels=True):
        r"""
        Initialize Square

        :param start_vertex: the vertex to start to draw the Square
        (default (0; 0))
        :type start_vertex: Point
        :param name: the name of the Square, like ABCDE for a pentagon. Can
        be either None (the names will be automatically created), or a
        string of the letters to use to name the vertices. Only single letters
        are supported as Points' names so far (at Polygon's creation).
        See issue #3.
        :type name: None or str
        :param side_length: the length that will be used to calculate the
        coordinates of the vertices used to build the Square
        :type side_length: a number
        :param mark_right_angles: if True (default), all four angles will be
        automatically marked as right angles.
        :type mark_right_angles: bool
        :param mark_equal_sides: if True (default), all four sides will be
        automatically marked with the same symbol.
        :type mark_equal_sides: bool
        :param draw_vertices: whether to actually draw, or not, the vertices
        :type draw_vertices: bool
        :param label_vertices: whether to label, or not, the vertices
        :type label_vertices: bool
        :param thickness: the thickness of the Polygon's sides
        :type thickness: str
        :param color: the color of the Polygon's sides
        :type color: str
        :param rotate: the angle of rotation around isobarycenter
        :type rotate: int
        """
        Rectangle.__init__(self, start_vertex=start_vertex, name=name,
                           width=side_length, length=side_length,
                           mark_right_angles=mark_right_angles,
                           draw_vertices=draw_vertices,
                           label_vertices=label_vertices,
                           thickness=thickness, color=color,
                           rotation_angle=rotation_angle,
                           winding=winding,
                           sloped_sides_labels=sloped_sides_labels)
        # TODO: see issue #5
        # Accepted type for side_length is number, already checked at
        # vertices' instanciations (in Rectangle.__init__() call).
        self._side_length = Number(side_length)
        Equilateral.__init__(self, mark_equal_sides=mark_equal_sides,
                             use_mark=use_mark)
        self._type = 'Square'

    @property
    def side_length(self):
        return self._side_length

    def setup_labels(self, labels=None, linesegments=None, masks=None):
        """Same as Rhombus.setup_labels()."""
        Rhombus.setup_labels(self, labels=labels, linesegments=linesegments,
                             masks=masks)

    @property
    def lbl_width(self):
        return self.lbl_side_length

    @property
    def lbl_length(self):
        return self.lbl_side_length



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from .equilateral import Equilateral
from .polygon import Polygon, shoelace_formula
from .triangle import Triangle
from .righttriangle import RightTriangle
from .equilateraltriangle import EquilateralTriangle
from .isoscelestriangle import IsoscelesTriangle
from .quadrilateral import Quadrilateral
from .rectangle import Rectangle
from .rhombus import Rhombus
from .square import Square

__all__ = ['Equilateral', 'Polygon', 'shoelace_formula',
           'Triangle', 'RightTriangle', 'EquilateralTriangle',
           'IsoscelesTriangle',
           'Quadrilateral', 'Rectangle', 'Rhombus', 'Square']



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import string
from decimal import InvalidOperation
from math import cos, sin, radians

from mathmakerlib import config
from mathmakerlib.core.drawable import Drawable, check_scale, tikz_options_list
from mathmakerlib.core.dimensional import Dimensional
from mathmakerlib.calculus.number import Number
from mathmakerlib.calculus.tools import is_number

OPPOSITE_LABEL_POSITIONS = {'right': 'left',
                            'above right': 'below left',
                            'above': 'below',
                            'above left': 'below right',
                            'left': 'right',
                            'below left': 'above right',
                            'below': 'above',
                            'below right': 'above left'}


class Point(Drawable, Dimensional):
    names_in_use = set()

    @classmethod
    def automatic_names(cls):
        use = Point.names_in_use
        names = [letter
                 for letter in string.ascii_uppercase
                 if letter not in use][::-1]
        if not names:
            layer = 1
            while not names:
                names = ['{}$_{}$'.format(letter, layer)
                         for letter in string.ascii_uppercase
                         if '{}$_{}$'.format(letter, layer)
                         not in use][::-1]
                layer += 1
        return names

    @classmethod
    def reset_names(cls):
        cls.names_in_use = set()

    def __init__(self, x=None, y=None, z='undefined', name='automatic',
                 color=None, shape=r'$\times$', shape_scale=Number('0.67'),
                 label='default', label_position='below'):
        r"""
        Initialize Point

        When not naming the keyword arguments, it is possible to not mention
        the applicate, z:
        Point(3, 2, 'A') is almost equivalent to Point(3, 2, 0, 'A'):
        - all subsequent keyword arguments must be named
        - the first version is a 2D Point, the second version a 3D Point.

        :param x: the Point's abscissa
        :type x: anything that can be turned to a Number
        :param y: the Point's ordinate
        :type y: anything that can be turned to a Number
        :param z: the Point's applicate
        :type z: anything that can be turned to a Number
        :param name: the Point's name (e.g. 'A'). If it's left to 'automatic',
        a yet unused name will be set.
        :type name: str
        :param shape: the symbol that will be drawn at the Point's position.
        Default value is '$\times$', what draws a cross.
        :type shape: str
        :param label: what will be placed near the Point. The default value
        will set the Point's name as label. Setting label at '' will disable
        labeling the Point.
        :type label: str
        :param label_position: if any label is to be drawn, this is where to
        draw it. Available values are TikZ's ones ('above', 'below left'...).
        :type label_position: str
        """
        self._three_dimensional = False
        self._name = None
        self._x = None
        self._y = None
        self._z = None
        self._shape = None
        self._label = None
        self._label_position = None
        self.x = x
        self.y = y
        try:
            self.z = z
        except (TypeError, InvalidOperation):
            # The third value cannot be used as z; let's assume it is actually
            # 2D geometry and the value is actually simply the name, implying
            # z should be set to 0.
            self.z = Number(0)
            self._three_dimensional = False
            self.name = z
        else:
            self.name = name
        self.shape = shape
        if label == 'default':
            self.label = self.name
        else:
            self.label = label
        self.label_position = label_position
        if color is not None:
            self.color = color
        self.shape_scale = shape_scale

    def __str__(self):
        if not self.three_dimensional:
            s = '{}({}, {})'.format(self.name, self.x, self.y)
        else:
            s = '{}({}, {}, {})'.format(self.name, self.x, self.y, self.z)
        return s

    def __repr__(self):
        if not self.three_dimensional:
            s = 'Point {}({}, {})'.format(self.name, self.x, self.y)
        else:
            s = 'Point {}({}, {}, {})'.format(self.name,
                                              self.x, self.y, self.z)
        return s

    def __hash__(self):
        if not self.three_dimensional:
            s = 'Point ({}, {})'.format(self.x, self.y)
        else:
            s = 'Point ({}, {}, {})'.format(self.x, self.y, self.z)
        return hash(s)

    def __eq__(self, other):
        if isinstance(other, Point):
            p = config.points.DEFAULT_POSITION_PRECISION
            return all([self.x.rounded(p) == other.x.rounded(p),
                        self.y.rounded(p) == other.y.rounded(p),
                        self.z.rounded(p) == other.z.rounded(p)])
        else:
            return False

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if self._name is not None:
            Point.names_in_use.discard(self._name)
        if value is None:
            self._name = None
        else:
            value = str(value)
            if value == 'automatic':
                self._name = Point.automatic_names().pop()
            else:
                self._name = value
            Point.names_in_use.add(self._name)

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, value):
        try:
            self._x = Number(value)
        except (TypeError, InvalidOperation):
            raise TypeError('Expected a number as abscissa, found {} '
                            'instead.'.format(repr(value)))

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        try:
            self._y = Number(value)
        except (TypeError, InvalidOperation):
            raise TypeError('Expected a number as ordinate, found {} '
                            'instead.'.format(repr(value)))

    @property
    def z(self):
        return self._z

    @z.setter
    def z(self, value):
        three_dimensional = True
        if value == 'undefined':
            value = 0
            three_dimensional = False
        try:
            self._z = Number(value)
        except (TypeError, InvalidOperation):
            raise TypeError('Expected a number as applicate, found {} '
                            'instead.'.format(repr(value)))
        self._three_dimensional = three_dimensional

    @property
    def coordinates(self):
        return (self._x, self._y, self._z)

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, other):
        self._shape = str(other)

    @property
    def shape_scale(self):
        return self._shape_scale

    @shape_scale.setter
    def shape_scale(self, value):
        check_scale(value, 'Point\'s shape')
        self._shape_scale = Number(value)

    @property
    def label_position(self):
        return self._label_position

    @label_position.setter
    def label_position(self, other):
        if other is None:
            self._label_position = None
        else:
            self._label_position = str(other)

    def rotate(self, center, angle, axis=None, rename='auto'):
        """
        Rotate around center (or axis going through center).

        :param center: the center of the rotation
        :type center: Point
        :param angle: the angle of the rotation
        :type angle: a number
        :param axis: the axis of the rotation for 3D rotation. If left to None,
        the rotation happens around the center, in the plane.
        :type axis: Vector
        :param rename: if set to 'auto', will name the rotated Point after the
        original, adding a ' (like A.rotate(...) creates a Point A'). If set
        to None, keep the original name. Otherwise, the provided str will be
        used as the rotated Point's name.
        :type rename: None or str
        :rtype: Point
        """
        from mathmakerlib.geometry.vector import Vector
        if not isinstance(center, Point):
            raise TypeError('Expected a Point as rotation center, got {} '
                            'instead.'.format(type(center)))
        if not is_number(angle):
            raise TypeError('Expected a number as rotation angle, got {} '
                            'instead.'.format(type(angle)))
        if not (axis is None or isinstance(axis, Vector)):
            raise TypeError('Expected either None or a Vector as axis, '
                            'found {} instead.'.format(repr(axis)))
        Δx = self.x - center.x
        Δy = self.y - center.y
        Δz = self.z - center.z
        cosθ = Number(str(cos(radians(angle))))
        sinθ = Number(str(sin(radians(angle))))
        if axis is None:
            rx = (Δx * cosθ - Δy * sinθ + center.x).rounded(Number('1.000'))
            ry = (Δx * sinθ + Δy * cosθ + center.y).rounded(Number('1.000'))
            rz = 'undefined'
        else:
            ux, uy, uz = axis.normalized().coordinates
            rotation_matrix = [[cosθ + (1 - cosθ) * ux ** 2,
                                ux * uy * (1 - cosθ) - uz * sinθ,
                                ux * uz * (1 - cosθ) + uy * sinθ],
                               [uy * ux * (1 - cosθ) + uz * sinθ,
                                cosθ + (1 - cosθ) * uy ** 2,
                                uy * uz * (1 - cosθ) - ux * sinθ],
                               [uz * ux * (1 - cosθ) - uy * sinθ,
                                uz * uy * (1 - cosθ) + ux * sinθ,
                                cosθ + (1 - cosθ) * uz ** 2]]
            rx = sum([rc * coord
                      for rc, coord in zip(rotation_matrix[0], [Δx, Δy, Δz])])
            ry = sum([rc * coord
                      for rc, coord in zip(rotation_matrix[1], [Δx, Δy, Δz])])
            rz = sum([rc * coord
                      for rc, coord in zip(rotation_matrix[2], [Δx, Δy, Δz])])
            rx += center.x
            ry += center.y
            rz += center.z
            rx = rx.rounded(Number('1.000'))
            ry = ry.rounded(Number('1.000'))
            rz = rz.rounded(Number('1.000'))

        if rename == 'keep_name':
            rname = self.name
        elif rename == 'auto':
            rname = self.name + "'"
        else:
            rname = rename
        return Point(rx, ry, rz, rname)

    def belongs_to(self, other):
        """Check if the Point belongs to a LineSegment."""
        # This could be extended to Lines and Rays, later.
        # Then check if self is different from the 2 known Points, and
        # if it belongs to the Line, then:
        # (x - x1) / (x2 - x1) = (y - y1) / (y2 - y1)
        # also, if three_dimensional, all this = (z - z1) / (z2 - z1)
        from mathmakerlib.geometry import LineSegment
        if not isinstance(other, LineSegment):
            raise TypeError('Argument \'other\' must be a LineSegment. '
                            'Found {} instead.'.format(repr(other)))
        d1 = LineSegment(self, other.endpoints[0]).length
        d2 = LineSegment(self, other.endpoints[1]).length
        return d1 + d2 == other.length

    def tikz_declaring_comment(self):
        """
        Replace plural by singular in the declaring comment.

        :rtype: str
        """
        return '% Declare Point'

    def tikz_declarations(self):
        """Return the Point declaration."""
        if self.name is None:
            raise RuntimeError('Point at ({}, {}) has no name (None), '
                               'cannot create TikZ picture using it.')
        return r'\coordinate ({}) at ({},{});'\
            .format(self.name,
                    self.x.rounded(Number('0.001')),
                    self.y.rounded(Number('0.001')))

    def tikz_drawing_comment(self):
        """Return the comment preceding the Point's drawing."""
        return ['% Draw Point']

    def _tikz_draw_options(self):
        return [self.color]

    def tikz_draw(self):
        """Return the command to actually draw the Point."""
        sh_scale = ''
        if self.shape_scale != 1:
            sh_scale = '[scale={}]'.format(self.shape_scale)
        return [r'\draw{} ({}) node{} {};'
                .format(tikz_options_list('draw', self),
                        self.name,
                        sh_scale,
                        '{' + self.shape + '}')]

    def tikz_labeling_comment(self):
        """
        Replace plural by singular in the labeling comment.

        :rtype: str
        """
        return '% Label Point'

    def tikz_points_labels(self):
        return self.tikz_label()

    def tikz_label(self):
        """Return the command to write the Point's label."""
        if self.label is None:
            return ''
        else:
            if self.label_position is None:
                label_position = ''
            else:
                label_position = '[' + self.label_position + ']'
            return r'\draw ({}) node{} {};'\
                .format(self.label, label_position, '{' + self.label + '}')



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from pathlib import Path

from mathmakerlib import required
from mathmakerlib.calculus.number import Number
from mathmakerlib.calculus.fraction import Fraction
from mathmakerlib.core.drawable import Drawable, HasThickness
from mathmakerlib.geometry.point import Point
from mathmakerlib.geometry.bipoint import Bipoint


class XAxis(Drawable, HasThickness, Bipoint):

    def __init__(self, mini, maxi, allow_zero_length=False, length=8, step=1,
                 subdivisions=1, before=None, after=2, thickness='thick',
                 points_def=None):
        """
        Initialize XAxis.

        :param mini: the value of the first shown axis' graduation
        :type mini: any Number (int or whatever can be fed to Number())
        :param maxi: the value of the last shown axis' graduation
        :type maxi: any Number (int or whatever can be fed to Number())
        :param thickness: the XAxis' thickness. Available values are TikZ's
        ones.
        :type thickness: str
        :param step: the value between two main graduations
        :type step: any Number (int or whatever can be fed to Number())
        :param subdivisions: number of subdivisions between the main
        graduations. 1 means no subdivisions, 2 means the main graduations will
        be divided in halves, 3 in thirds etc.
        :type subdvisions: int >= 1
        :param before: number of subdivisions ahead of the first main
        graduation.
        :type before: int
        :param after: number of subdivisions after the first main graduation
        :type after: int
        :param length: the length of the axis
        :type length: any Number (int or whatever can be fed to Number())
        :param allow_zero_length: whether creation of a zero length axis is
        allowed or not (default False)
        :type allow_zero_length: bool
        :param points_def: a list of pairs (abscissa, name) of the points to be
        drawn on the axis. The abscissae to provide must match points of the
        axis. If an abscissa is outside [mini - before, maxi + after], it
        will be dropped with a warning.
        :type points_def: None or a list
        """
        mini = Number(mini)
        maxi = Number(maxi)
        step = Number(step)
        length = Number(length)
        # extra graduation that won't be drawn, to leave some space
        # between the last drawn graduation and the arrow
        after += 1
        endpoints = [Point(x=0, y=0), Point(x=length, y=0)]
        Bipoint.__init__(self, *endpoints, allow_zero_length=allow_zero_length)
        if before is None:
            if mini == 0:
                before = 0
            else:  # mini > 0
                if subdivisions in [1, 2]:
                    before = 2
                else:  # subdivisions >= 3
                    before = 3
        # mg = main graduations
        # sg = sub graduations
        mg_number = int((maxi - mini) / step)
        sg_number = int(before + mg_number * subdivisions + after)
        all_sg = self.dividing_points(n=sg_number)
        self._sg_abscissae = [p.x.rounded(Number('1.00')) for p in all_sg]
        sub_multiples = [i for i in range(sg_number) if not (i % subdivisions)]
        mg_indices = [n + before - 1 for n in sub_multiples]
        mg_indices = [i for i in mg_indices if i < len(all_sg)]
        if before == 0:
            all_sg.append(Point(0, 0))
        self._mg_abscissae = [all_sg[i].x.rounded(Number('1.00'))
                              for i in mg_indices]
        self._mg_labels = [Number(mini + n * step).printed
                           for n in range(mg_number + 1)]
        if points_def is None:
            points_def = []
        sg_step = step / subdivisions
        m = mini - before * sg_step
        # TODO: check there are no abscissae outside the drawn part of the axis
        abscissae = [x.evaluate()
                     if isinstance(x, Fraction) else Number(x).evaluate()
                     for (x, _) in points_def]
        abscissae = [(x.evaluate() - m) * length / (sg_step * sg_number)
                     for x in abscissae]
        names = [n for (_, n) in points_def]
        points_def = zip(abscissae, names)
        self._points = [Point(x=abscissa.rounded(Number('1.00')), y=0, name=n,
                              label_position='above')
                        for (abscissa, n) in points_def]
        self._length = Number(length)

    @property
    def template_fmt(self):
        sg_abscissae = ','.join([a.imprint(dot=True)
                                 for a in self._sg_abscissae])
        # do not leave commas in the labels to be printed, it will mess up
        # the axis (because TiKZ already uses commas to separate the values)
        if any([',' in label for label in self._mg_labels]):
            mg_abscissae_texts = \
                ','.join([f'{x}/{{{lab}}}'
                          for (x, lab) in zip(self._mg_abscissae,
                                              self._mg_labels)])
        else:
            mg_abscissae_texts = \
                ','.join([f'{x}/{lab}' for (x, lab) in zip(self._mg_abscissae,
                                                           self._mg_labels)])
        points_drawn = [(r'\draw[thick] ({x},0) node {{$\times$}} '
                         r'node[{lbl_pos}] {{{name}}};')
                        .format(x=p.x, lbl_pos=p.label_position, name=p.name)
                        for p in self._points]
        return {'__TIKZ_PICTURE_OPTIONS__': self.tikz_picture_options(),
                '__LENGTH__': self._length.imprint(dot=True),
                '__SG_ABSCISSAE__': '{' + sg_abscissae + '}',
                '__MG_ABSCISSAE_TEXTS__': '{' + mg_abscissae_texts + '}',
                '__POINTS_DRAWN__': '\n'.join(points_drawn)
                }

    def draw(self):
        required.package['tikz'] = True
        pic = (Path(__file__).parent / 'templates/xaxis.tex').read_text()
        for placeholder in self.template_fmt:
            pic = pic.replace(placeholder, self.template_fmt[placeholder])
        return pic

    def _tikz_draw_options(self):
        pass

    def tikz_declarations(self):
        pass

    def tikz_draw(self):
        pass

    def tikz_drawing_comment(self):
        pass

    def tikz_label(self):
        return ''

    def tikz_points_labels(self):
        pass



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import warnings
from copy import deepcopy
from statistics import mean

from mathmakerlib import config
from mathmakerlib.calculus.number import Number
from mathmakerlib.calculus.tools import is_number
from mathmakerlib.core.drawable import Drawable, HasThickness, Colored
from mathmakerlib.core.oriented import Oriented, check_winding
from mathmakerlib.core.oriented import shoelace_formula
from mathmakerlib.core.drawable import tikz_approx_position, tikz_options_list
from mathmakerlib.core.dimensional import Dimensional
from mathmakerlib.geometry.point import Point
from mathmakerlib.geometry.linesegment import LineSegment
from mathmakerlib.geometry.vector import Vector
from mathmakerlib.geometry.angle import Angle

POLYGONS_TYPES = {3: 'Triangle', 4: 'Quadrilateral', 5: 'Pentagon',
                  6: 'Hexagon', 7: 'Heptagon', 8: 'Octogon',
                  9: 'Nonagon', 10: 'Decagon', 11: 'Hendecagon',
                  12: 'Dodecagon', 13: 'Tridecagon', 14: 'Tetradecagon',
                  15: 'Pentadecagon', 16: 'Hexadecagon', 17: 'Heptadecagon',
                  18: 'Octodecagon', 19: 'Enneadecagon', 20: 'Icosagon'}


class Polygon(Drawable, Colored, HasThickness, Oriented, Dimensional):
    """Polygons."""

    def __init__(self, *vertices, name=None,
                 draw_vertices=False, label_vertices=True,
                 thickness='thick', color=None, rotation_angle=0,
                 winding=None, do_cycle=True, sloped_sides_labels=True):
        r"""
        Initialize Polygon

        :param vertices: the vertices of the Polygon
        :type vertices: a list of at least three Points
        :param name: the name of the Polygon, like ABCDE for a pentagon. Can
        be either None (the names of the provided Points will be kept), or a
        string of the letters to use to rename the provided Points. Only
        single letters are supported as Points' names so far (at Polygon's
        creation). See issue #3.
        :type name: None or str
        :param draw_vertices: whether to actually draw, or not, the vertices
        :type draw_vertices: bool
        :param label_vertices: whether to label, or not, the vertices
        :type label_vertices: bool
        :param thickness: the thickness of the Polygon's sides
        :type thickness: str
        :param color: the color of the Polygon's sides
        :type color: str
        :param rotate: the angle of rotation around isobarycenter
        :type rotate: int
        :param winding: force the winding to be either 'clockwise' or
        'anticlockwise'. If left to None (default), doesn't force anything,
        the winding will be either forced by the value of
        config.DEFAULT_POLYGON_WINDING, or if it is None too, then the
        winding will be deduced from the given vertices' order.
        :type winding: None or a str ('clockwise' or 'anticlockwise')
        """
        self.thickness = thickness
        self.color = color
        self.do_cycle = do_cycle
        self.draw_vertices = draw_vertices
        self.label_vertices = label_vertices
        if len(vertices) <= 2:
            raise ValueError('At least three Points are required to be able '
                             'to build a Polygon. Got only {} positional '
                             'arguments, though.'.format(len(vertices)))
        if any([not isinstance(v, Point) for v in vertices]):
            for i, v in enumerate(vertices):
                if not isinstance(v, Point):
                    raise TypeError('Only Points must be provided in order to '
                                    'build a Polygon. Got a {} as positional '
                                    'argument #{}.'.format(type(v), i))
        if not is_number(rotation_angle):
            raise TypeError('Expected a number as rotation angle, got a {} '
                            'instead.'.format(type(rotation_angle)))
        if name is not None:
            if len(name) != len(vertices):
                raise ValueError('The number of provided vertices ({}) does '
                                 'not match the number of Points\' names '
                                 '({}).'.format(len(vertices), len(name)))

        if (winding is None
            and config.polygons.DEFAULT_WINDING is not None):
            winding = config.polygons.DEFAULT_WINDING

        if winding is not None:
            check_winding(winding)

        self._reverted_winding = False

        if shoelace_formula(*vertices) < 0:
            if winding == 'anticlockwise':
                vertices = vertices[::-1]
                self._reverted_winding = True
                self.winding = 'anticlockwise'
            else:
                self.winding = 'clockwise'
        else:
            if winding == 'clockwise':
                vertices = vertices[::-1]
                self._reverted_winding = True
                self.winding = 'clockwise'
            else:
                self.winding = 'anticlockwise'

        self._vertices = []
        self._three_dimensional = False
        for i, v in enumerate(vertices):
            if name is None:
                vname = v.name
            else:
                vname = name[i]
            if v.name == v.label:
                lbl = 'default'
            else:
                lbl = v.label
            if v.three_dimensional:
                self._three_dimensional = True
                zval = v.z
            else:
                zval = 'undefined'
            self._vertices.append(Point(v.x, v.y, z=zval, name=vname,
                                        shape=v.shape,
                                        label=lbl, color=v.color,
                                        shape_scale=v.shape_scale))
        if rotation_angle:
            center = self.isobarycenter()
            for i in range(len(self._vertices)):
                self._vertices[i] = self._vertices[i].rotate(
                    center=center,
                    angle=rotation_angle,
                    rename='keep_name')

        self._sides = []
        shifted_vertices = deepcopy(self._vertices)
        shifted_vertices += [shifted_vertices.pop(0)]
        for (v0, v1) in zip(self._vertices, shifted_vertices):
            self._sides += [LineSegment(v0, v1, label_winding=self.winding,
                                        locked_label=True)]
        self._angles = []
        left_shifted_vertices = deepcopy(self._vertices)
        left_shifted_vertices = \
            [left_shifted_vertices.pop(-1)] + left_shifted_vertices
        for (v0, v1, v2) in zip(left_shifted_vertices,
                                self._vertices,
                                shifted_vertices):
            self._angles += [Angle(v2, v1, v0)]
        for i in range(len(self._vertices)):
            u = Vector(self._vertices[i], left_shifted_vertices[i])
            v = Vector(self._vertices[i], shifted_vertices[i])
            if self.winding == 'clockwise':
                u, v = v, u
            self._vertices[i].label_position = \
                tikz_approx_position(u.bisector(v).slope360)

        if len(self._sides) in POLYGONS_TYPES:
            self._type = POLYGONS_TYPES[len(self._sides)]
        else:
            self._type = \
                '{n}-sided Polygon'.format(n=str(len(self._sides)))

        self.sloped_sides_labels = sloped_sides_labels

        if (self._reverted_winding
            and config.polygons.ENABLE_MISMATCH_WINDING_WARNING):
            warnings.warn('Changed the order of Points to comply with forced '
                          'winding ({}) for {}.'.format(winding, repr(self)))

    def __repr__(self):
        return '{} {}'.format(self.type, self.name)

    def __eq__(self, other):
        if isinstance(other, Polygon) and len(self.sides) == len(other.sides):
            L = len(self.vertices)
            for offset in range(L):
                if all([self.vertices[(i + offset) % L] == other.vertices[i]
                        for i in range(L)]):
                    return True
            for offset in range(L):
                if all([self.vertices[(i + offset) % L] == other.vertices[-i]
                        for i in range(L)]):
                    return True
        return False

    @property
    def vertices(self):
        return self._vertices

    @property
    def sides(self):
        return self._sides

    @property
    def angles(self):
        return self._angles

    @property
    def name(self):
        return ''.join([v.name for v in self.vertices])

    @property
    def type(self):
        return self._type

    def isobarycenter(self, name='automatic'):
        if self.three_dimensional:
            zval = mean([v.z for v in self.vertices])
        else:
            zval = 'undefined'
        return Point(mean([v.x for v in self.vertices]),
                     mean([v.y for v in self.vertices]),
                     z=zval,
                     name=name)

    @property
    def perimeter(self):
        return sum([s.length for s in self.sides])

    @property
    def lbl_perimeter(self):
        if any([not isinstance(s.label_value, Number) for s in self.sides]):
            raise RuntimeError('All labels must have been set as Numbers '
                               'in order to calculate the perimeter from '
                               'labels.')
        else:
            return sum([s.label_value for s in self.sides],
                       Number(0, unit=self.sides[0].label_value.unit))\
                .standardized()

    def setup_labels(self, labels=None, linesegments=None, masks=None):
        """
        Convenience method to easily setup all sides' labels in a row.

        If labels is None, then the labels won't be set.
        If masks is None, then the masks won't be set.
        At least labels or masks must be different from None.

        If linesegments is None, it defaults to the Polygon's sides. This is
        practical if extra line segments require labeling, like in
        InterceptTheoremFigure, or if one wants to label a diagonal.

        It is expected that both the labels' and line segments' lists
        have the same length.

        :param labels: None or the list of the labels
        :type labels: list
        :param linesegments: the list of the LineSegments to label
        (defaults to Polygon's sides)
        :type linesegments: list (of LineSegments)
        :param masks: the list of masks to setup. If None (default), all masks
        will be set to None.
        :type masks: None or list
        """
        if labels is None and masks is None:
            raise ValueError('There must be at least either labels or masks '
                             'to setup. Both are undefined (None).')
        if linesegments is None:
            linesegments = self.sides
        if labels is not None and len(labels) != len(linesegments):
            raise ValueError('The number of labels ({}) should be equal '
                             'to the number of line segments ({}).'
                             .format(str(len(labels)),
                                     str(len(linesegments))))
        if masks is None:
            masks = [None for _ in range(len(linesegments))]
        if len(masks) != len(linesegments):
            raise ValueError('The number of label masks ({}) should be equal '
                             'to the number of line segments ({}).'
                             .format(str(len(masks)),
                                     str(len(linesegments))))
        if (labels is not None and self._reverted_winding
            and linesegments == self.sides):
            labels = labels[::-1]
            labels += [labels.pop(0)]
        if labels is not None:
            for (ls, lbl) in zip(linesegments, labels):
                ls.unlock_label()
                ls.label = lbl
                ls.lock_label()
        for (ls, m) in zip(linesegments, masks):
            ls.label_mask = m

    def setup_marks(self, marks=None, linesegments=None):
        """
        Convenience method to easily setup all sides' marks in a row.

        If marks is None, no LineSegment's mark will be set.

        :param marks: None or the list of the marks
        :type marks: list
        :param linesegments: the list of the LineSegments to mark
        (defaults to Polygon's sides)
        :type linesegments: list (of LineSegments)
        """
        if linesegments is None:
            linesegments = self.sides
        if (marks is not None and self._reverted_winding
            and linesegments == self.sides):
            marks = marks[::-1]
            marks += [marks.pop(0)]
        if marks is not None and len(marks) == len(linesegments):
            for s, m in zip(linesegments, marks):
                s.mark = m

    @property
    def sloped_sides_labels(self):
        return self._sloped_sides_labels

    @sloped_sides_labels.setter
    def sloped_sides_labels(self, value):
        if isinstance(value, bool):
            self._sloped_sides_labels = value
            for s in self.sides:
                s.sloped_label = value
                s.label_position = 'automatic'
        else:
            raise TypeError('sloped_sides_labels must be a boolean; '
                            'got {} instead.'.format(type(value)))

    @property
    def do_cycle(self):
        return self._do_cycle

    @do_cycle.setter
    def do_cycle(self, value):
        if isinstance(value, bool):
            self._do_cycle = value
        else:
            raise TypeError('do_cycle must be a boolean; '
                            'got {} instead.'.format(type(value)))

    @property
    def draw_vertices(self):
        return self._draw_vertices

    @draw_vertices.setter
    def draw_vertices(self, value):
        if isinstance(value, bool):
            self._draw_vertices = value
        else:
            raise TypeError('draw_vertices must be a boolean; '
                            'got {} instead.'.format(type(value)))

    @property
    def label_vertices(self):
        return self._label_vertices

    @label_vertices.setter
    def label_vertices(self, value):
        if isinstance(value, bool):
            self._label_vertices = value
        else:
            raise TypeError('label_vertices must be a boolean; '
                            'got {} instead.'.format(type(value)))

    def tikz_declarations(self):
        """Return the Points declarations."""
        return '\n'.join([v.tikz_declarations() for v in self.vertices])

    def tikz_drawing_comment(self):
        """Return the comment preceding the Polygon's drawing."""
        output = []
        if self.draw_vertices:
            output.append('% Draw Vertices')
        output.append('% Draw {}'.format(self.type))
        if any([θ.mark_right for θ in self.angles]):
            output.append('\n% Mark right angles')
        return output

    def _tikz_draw_options(self):
        return [self.thickness, self.color]

    def _tikz_draw_vertices(self):
        return '\n'.join([v.tikz_draw()[0] for v in self.vertices]) + '\n'

    def _tikz_draw_sides(self):
        if self.do_cycle:
            return '\n'.join([s.tikz_draw_section() for s in self.sides[:-1]])
        else:
            return '\n'.join([s.tikz_draw_section() for s in self.sides])

    def _tikz_draw_angles_marks(self):
        marks = '\n'.join([a.tikz_decorations()
                          for a in self.angles
                          if a.tikz_decorations() != ''])
        if marks == '':
            return ''
        else:
            return '\n' + marks

    def _tikz_draw_right_angles_marks(self):
        return '\n'.join([θ.tikz_rightangle_mark(self.winding)
                          for θ in self.angles
                          if θ.tikz_rightangle_mark(self.winding) != ''])

    def tikz_draw(self):
        """Return the command to actually draw the Polygon."""
        output = []
        if self.draw_vertices:
            output.append(self._tikz_draw_vertices())
        if self.do_cycle:
            output.append(r'\draw' + '{} ({})\n{}\n-- cycle{}{}{};'
                          .format(tikz_options_list('draw', self),
                                  self.vertices[0].name,
                                  self._tikz_draw_sides(),
                                  self.sides[-1].tikz_label(),
                                  self.sides[-1]._tikz_ls_mark(),
                                  self._tikz_draw_angles_marks()))
        else:
            output.append(r'\draw' + '{} ({})\n{}{} -- cycle;'
                          .format(tikz_options_list('draw', self),
                                  self.vertices[0].name,
                                  self._tikz_draw_sides(),
                                  self._tikz_draw_angles_marks()))
        if any([θ.mark_right for θ in self.angles]):
            output.append(self._tikz_draw_right_angles_marks())
        return output

    def tikz_points_labels(self):
        """Return the command to write the Points' labels."""
        if self.label_vertices:
            return '\n'.join([v.tikz_label() for v in self.vertices])
        return ''

    def tikz_label(self):
        """Not implemented yet. See issue #4."""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mathmakerlib',
 'mathmakerlib.LaTeX',
 'mathmakerlib.calculus',
 'mathmakerlib.core',
 'mathmakerlib.geometry',
 'mathmakerlib.geometry.polygons',
 'mathmakerlib.geometry.polyhedra',
 'mathmakerlib.geometry.projections']

package_data = \
{'': ['*'], 'mathmakerlib.geometry': ['templates/*']}

install_requires = \
['toml>=0.10.2,<0.11.0']

setup_kwargs = {
    'name': 'mathmakerlib',
    'version': '0.7.11',
    'description': 'Collection of various useful tools.',
    'long_description': '|PyPI1| |PyPI2| |PyPI3| |Codecov| |Build Status| |Documentation Status| |Maintenance|\n\n|PyPI4|\n\n\nOverview\n========\n\nMathmaker Lib offers lualatex-printable mathematical objects.\n\n`Repo is here <https://gitlab.com/nicolas.hainaux/mathmakerlib>`__\n\n`See documentation here <http://mathmaker-lib.readthedocs.io/>`__\n\nContact: nh dot techn (hosted by gmail dot com)\n\n.. |PyPI1| image:: https://img.shields.io/pypi/v/mathmakerlib.svg?maxAge=2592000\n   :target: https://pypi.python.org/pypi/mathmakerlib\n.. |PyPI2| image:: https://img.shields.io/pypi/status/mathmakerlib.svg?maxAge=2592000\n.. |PyPI3| image:: https://img.shields.io/pypi/pyversions/mathmakerlib.svg?maxAge=2592000\n.. |Build Status| image:: https://ci.appveyor.com/api/projects/status/7vejgm0hjm6236xo/branch/master?svg=true\n   :target: https://ci.appveyor.com/project/nicolashainaux/mathmakerlib-ho94f\n.. |Codecov| image:: https://codecov.io/gl/nicolas.hainaux/mathmakerlib/branch/master/graph/badge.svg\n  :target: https://codecov.io/gl/nicolas.hainaux/mathmakerlib\n.. |Documentation Status| image:: https://readthedocs.org/projects/mathmaker-lib/badge/?version=latest\n   :target: https://mathmaker-lib.readthedocs.io/en/latest/\n.. |Maintenance| image:: https://img.shields.io/maintenance/yes/2019.svg?maxAge=2592000\n.. |PyPI4| image:: https://img.shields.io/pypi/l/mathmakerlib.svg?maxAge=2592000\n   :target: https://gitlab.com/nicolas.hainaux/mathmakerlib/blob/master/LICENSE\n',
    'author': 'Nicolas Hainaux',
    'author_email': 'nh.techn@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from pathlib import Path

import toml

from . import required, exceptions, config, constants
from . import calculus, core, geometry, LaTeX

__all__ = ['required', 'config', 'LaTeX', 'exceptions', 'constants',
           'core', 'calculus', 'geometry']

with open(Path(__file__).parent.parent / 'pyproject.toml', 'r') as f:
    pp = toml.load(f)

__version__ = pp['tool']['poetry']['version']
__lib_name__ = pp['tool']['poetry']['name']

__release__ = __version__ + ' (alpha)'
__author__ = 'Nicolas Hainaux'
__author_email__ = 'nh.techn@gmail.com'
__licence__ = 'GNU General Public License v3 or later (GPLv3+)'
__url__ = 'https://gitlab.com/nicolas.hainaux/mathmakerlib/'
__copyright__ = 'Copyright 2006-2019'
__contact__ = '{author} <{author_email}>'\
              .format(author=__author__, author_email=__author_email__)
__licence_info__ = '{lib_ref} is free software. Its license is '\
                   '{lib_license}.'
__url_info__ = 'Further details on {lib_website}'
__info__ = '{lib_name} {r}\nLicense: {li}\n{c} {contact}'\
           .format(lib_name=__lib_name__,
                   r=__release__, li=__licence__, c=__copyright__,
                   contact=__contact__)

config.init()
required.init()



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


def init():
    global package, options, tikz_library, tikzset
    global required_initialized

    try:
        required_initialized
    except NameError:
        required_initialized = False

    if not required_initialized:
        required_initialized = True
        # It's difficult to track amssymb, that could show up almost anywhere.
        # This is left to mathmakerlib's user. A (yet short) list of symbols is
        # provided in LaTeX module.
        package = {pkg_name: False
                   for pkg_name in ['tikz', 'siunitx', 'xcolor', 'eurosym',
                                    'amsmath', 'stackengine', 'scalerel',
                                    'cancel', 'multicol', 'placeins', 'ulem',
                                    'textcomp', 'array', 'graphicx',
                                    'epstopdf', 'textpos', 'fancyvrb']}
        options = {'xcolor': set(), 'textpos': set()}
        tikz_library = {'angles': False,
                        'decorations.markings': False,
                        'quotes': False}
        tikzset = {'singledash_hatchmark': False,
                   'doubledash_hatchmark': False,
                   'tripledash_hatchmark': False}
        # hack = {'rightangle_mark': False}



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from decimal import Decimal

from mathmakerlib.core.drawable import check_color, tikz_options_list
from mathmakerlib.calculus.tools import is_number, is_integer
from mathmakerlib.calculus.number import Number
from mathmakerlib.geometry.linesegment import LineSegment


class DividedLineSegment(LineSegment):

    def __init__(self, *points, thickness='ultra thick', label=None,
                 label_mask=None,
                 label_position='anticlockwise',
                 draw_endpoints=True, label_endpoints=False, color=None,
                 n=None, inner_points_shape='|',
                 endpoints_shape_as_inner_points=True,
                 fill=0, fillcolor='LimeGreen'):
        """
        Initialize DividedLineSegment

        First parameters are the same as for LineSegment. Default values may
        differ, though.
        The endpoints will be set to the same shape as others (they must be
        modified after initialization if desired another way).
        """
        LineSegment.__init__(self, *points, thickness=thickness, label=label,
                             label_mask=label_mask,
                             label_position=label_position,
                             draw_endpoints=draw_endpoints,
                             label_endpoints=label_endpoints,
                             color=color)
        self.n = n
        self.fill = fill
        self.inner_points_shape = inner_points_shape
        if endpoints_shape_as_inner_points:
            self.endpoints[0].shape = inner_points_shape
            self.endpoints[1].shape = inner_points_shape
        self._comment_designation = 'Divided Line Segment'
        self.fillcolor = fillcolor

    def __repr__(self):
        return 'DividedLineSegment({}, {}, {}/{}, {})' \
            .format(repr(self.endpoints[0]),
                    repr(self.endpoints[1]),
                    self.fill, self.n,
                    self.fillcolor)

    @property
    def n(self):
        return self._n

    @n.setter
    def n(self, n):
        if not (is_number(n) and is_integer(n) and n >= 1):
            raise TypeError('n must be an integer >= 1, got {} instead.'
                            .format(n))
        self._n = Number(n)

    @property
    def fill(self):
        return self._fill

    @fill.setter
    def fill(self, value):
        if not (is_integer(value) and 1 <= value <= self.n):
            raise TypeError('fill must be an 1 <= integer <= self.n == {}, '
                            'got {} instead.'
                            .format(self.n, value))
        self._fill = Number(value)

    @property
    def inner_points_shape(self):
        return self._inner_points_shape

    @inner_points_shape.setter
    def inner_points_shape(self, other):
        self._inner_points_shape = str(other)

    @property
    def fillcolor(self):
        return self._fillcolor

    @fillcolor.setter
    def fillcolor(self, value):
        check_color(value)
        self._fillcolor = value

    @property
    def fillpoint(self):
        fp = ([self.endpoints[0]] + self.dividing_points(self.n)
              + [self.endpoints[1]])[int(self.fill)]
        fp.x = fp.x.rounded(Decimal('0.001'))
        fp.y = fp.y.rounded(Decimal('0.001'))
        return fp

    def tikz_graduations(self):
        """Return the DividedLineSegment's nodes' list."""
        points_list = [(((i + 1) / self.n).rounded(Decimal('0.001')),
                        self.inner_points_shape)
                       for i in range(int(self.n - 1))]
        if self.draw_endpoints:
            points_list = [(0, self.endpoints[0].shape)] + points_list
            points_list = points_list + [(1, self.endpoints[1].shape)]
        return ' '.join(['node[opacity=1, pos={}, sloped] {}'
                         .format(position, '{' + shape + '}')
                         for (position, shape) in points_list])

    def tikz_declarations(self):
        """Return the DividedLineSegment's Points' declarations."""
        return '{}\n{}\n{}'\
            .format(self.endpoints[0].tikz_declarations(),
                    self.endpoints[1].tikz_declarations(),
                    self.fillpoint.tikz_declarations())

    def tikz_drawing_comment(self):
        """Return the comments preceding the DividedLineSegment's drawings."""
        return ['% Draw {}'.format(self._comment_designation)]

    def tikz_draw(self):
        """Return the command to actually draw the DividedLineSegment."""
        output = []
        draw_cmd = r'''\draw{} ({}) -- ({}){};
\draw{} ({}) -- ({});
\draw{} ({}) -- ({}) {};'''\
    .format(tikz_options_list('draw', self),
            self.endpoints[0].name, self.endpoints[1].name,
            self.tikz_label(),
            tikz_options_list([self.thickness, self.fillcolor]),
            self.endpoints[0].name, self.fillpoint.name,
            tikz_options_list([self.thickness, 'opacity=0']),
            self.endpoints[0].name, self.endpoints[1].name,
            self.tikz_graduations())
        output.append(draw_cmd)
        return output



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


class MathmakerLibError(Exception):
    """Basic exception for errors raised by Mathmaker Lib."""
    def __init__(self, msg=None):
        if msg is None:
            # Set some default useful error message
            msg = 'An error occured in Mathmaker Lib'
        super().__init__(msg)


class StopCalculation(MathmakerLibError):
    """When no further calculation can be made."""
    def __init__(self, o, msg=None):
        if msg is None:
            msg = 'No further calculation can be done on {}.'.format(repr(o))
        super().__init__(msg=msg)


class StopFractionReduction(StopCalculation):
    """When no further reduction can be made."""
    def __init__(self, f, msg=None):
        super().__init__(f,
                         msg='{} can no further be reduced.'.format(repr(f)))


class ZeroBipoint(MathmakerLibError):
    """In case of abusive use of a zero-length Bipoint."""
    def __init__(self, msg=None):
        if msg is None:
            msg = 'Abusive use of a zero Bipoint.'
        super().__init__(msg=msg)


class ZeroVector(MathmakerLibError):
    """In case of abusive use of a zero-length Vector."""
    def __init__(self, msg=None):
        if msg is None:
            msg = 'Abusive use of a zero Vector.'
        super().__init__(msg=msg)


class ZeroLengthLineSegment(ZeroBipoint):
    """In case of abusive use of a zero-length LineSegment."""
    def __init__(self, msg=None):
        if msg is None:
            msg = 'Abusive use of a zero-length LineSegment.'
        super().__init__(msg=msg)


ZERO_OBJECTS_ERRORS = {'Bipoint': ZeroBipoint,
                       'LineSegment': ZeroLengthLineSegment,
                       'Vector': ZeroVector}



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from math import radians, sin, cos

from mathmakerlib import config
from mathmakerlib.core.drawable import Drawable, tikz_approx_position
from mathmakerlib.core.drawable import HasThickness
from mathmakerlib.geometry.tools import convex_hull
from mathmakerlib.calculus.tools import is_number
from mathmakerlib.geometry.point import Point
from mathmakerlib.geometry.linesegment import LineSegment
from mathmakerlib.geometry.vector import Vector
from mathmakerlib.geometry.polyhedra import Polyhedron

DIRECTION_VALUES = ['top-left', 'top-right', 'bottom-left', 'bottom-right']


def check_direction(value):
    if value not in DIRECTION_VALUES:
        raise ValueError('Allowed values for direction argument are {}. '
                         'Found {} instead.'
                         .format(DIRECTION_VALUES, repr(value)))


class ObliqueProjection(Drawable, HasThickness):
    """
    Oblique projection of 3D objects.

    The objects will be projected on the XOY plane.
    Introduced in version 0.7, supporting only Polyhedra projections yet.
    """

    def __init__(self, object3D=None, k=None, α=None, direction=None,
                 thickness='thick', color=None,
                 draw_vertices=False, label_vertices=False):
        r"""
        Initialize ObliqueProjection.

        :param object3D: the object to project.
        :type object3D: Polyhedron
        :param k: the ratio of the oblique projection. Defaults to
        config.oblique_projection.RATIO
        :type k: number
        :param α: the angle between the receding Z-axis, and X-axis.
        :type α: number
        :param direction: 'top-left', 'top-right', 'bottom-left' or
        'bottom-right'
        :type direction: str
        """
        self.draw_vertices = draw_vertices
        self.label_vertices = label_vertices
        if k is None:
            k = config.oblique_projection.RATIO
        if not is_number(k):
            raise TypeError('Ratio k must be a number. Found {} instead.'
                            .format(repr(k)))
        if α is None:
            α = config.oblique_projection.RECEDING_AXIS_ANGLE
        if direction is None:
            direction = config.oblique_projection.DIRECTION
        if not is_number(α):
            raise TypeError('Angle α must be a number. Found {} instead.'
                            .format(repr(α)))
        if not isinstance(object3D, Polyhedron):
            raise TypeError('object3D must be a Polyhedron, found {} instead.'
                            .format(repr(object3D)))
        check_direction(direction)
        self._direction = direction
        self._object3D_name = type(object3D).__name__

        # Setup the edges' labels
        if object3D.labels is not None:
            edges_to_label = object3D.edges_to_label['oblique_projection:{}'
                                                     .format(direction)]
            w_coord, d_coord, h_coord = edges_to_label
            object3D.faces[w_coord[0]].sides[w_coord[1]].unlock_label()
            object3D.faces[w_coord[0]].sides[w_coord[1]].label_winding = \
                w_coord[2]
            object3D.faces[w_coord[0]].sides[w_coord[1]].label = \
                object3D.labels[0]
            object3D.faces[w_coord[0]].sides[w_coord[1]].lock_label()
            object3D.faces[d_coord[0]].sides[d_coord[1]].unlock_label()
            object3D.faces[d_coord[0]].sides[d_coord[1]].label_winding = \
                d_coord[2]
            object3D.faces[d_coord[0]].sides[d_coord[1]].label = \
                object3D.labels[1]
            object3D.faces[d_coord[0]].sides[d_coord[1]].lock_label()
            object3D.faces[h_coord[0]].sides[h_coord[1]].unlock_label()
            object3D.faces[h_coord[0]].sides[h_coord[1]].label_winding = \
                h_coord[2]
            object3D.faces[h_coord[0]].sides[h_coord[1]].label = \
                object3D.labels[2]
            object3D.faces[h_coord[0]].sides[h_coord[1]].lock_label()

        matrix = {'top-right': [[1, 0, k * sin(radians(α))],
                                [0, 1, k * cos(radians(α))]],
                  'bottom-right': [[1, 0, k * sin(radians(α))],
                                   [0, 1, -k * cos(radians(α))]],
                  'bottom-left': [[1, 0, -k * sin(radians(α))],
                                  [0, 1, -k * cos(radians(α))]],
                  'top-left': [[1, 0, -k * sin(radians(α))],
                               [0, 1, k * cos(radians(α))]]}[direction]
        self._edges = []
        self._edges3D = {}
        self._vertices = []
        self._vertices_match = {}

        def project(point, matrix):
            x = sum([pc * coord
                     for pc, coord in zip(matrix[0],
                                          point.coordinates)])
            y = sum([pc * coord
                     for pc, coord in zip(matrix[1],
                                          point.coordinates)])
            return Point(x, y, point.name)

        for vertex in object3D.vertices:
            projected_point = project(vertex, matrix)
            self._vertices.append(projected_point)
            self._vertices_match[vertex.name] = (vertex, projected_point)

        # To store to which edges a vertex belongs
        vertices_connexions = {k: [] for k in self._vertices}

        # Build the projected edges
        for edge in object3D.edges:
            p0 = self._vertices_match[edge.endpoints[0].name][1]
            p1 = self._vertices_match[edge.endpoints[1].name][1]
            # TODO: check cases when the projected edge is a single point
            # (ZeroBipoint should be raised)
            projected_edge = LineSegment(p0, p1,
                                         thickness=thickness,
                                         draw_endpoints=draw_vertices,
                                         label_endpoints=label_vertices,
                                         color=color,
                                         allow_zero_length=False,
                                         locked_label=True,
                                         label_scale=edge.label_scale,
                                         label=edge.label,
                                         label_mask=edge.label_mask,
                                         label_winding=edge.label_winding,
                                         sloped_label=False)
            vertices_connexions[p0].append(LineSegment(p0, p1))
            vertices_connexions[p1].append(LineSegment(p1, p0))
            if projected_edge not in self._edges:  # TODO: else, what...?
                self._edges.append(projected_edge)
            if projected_edge not in self._edges3D:  # TODO: else, what...?
                self._edges3D[projected_edge] = edge
        # Find out which edges are hidden.
        # The ones that belong to convex hull of the projected vertices are
        # considered visible. By default, they will remain visible (i.e. keep
        # the default 'solid' dashpattern).
        points_cloud = set()  # to avoid duplicates
        for edge in self.edges:
            points_cloud.update(edge.endpoints)
        cvh = convex_hull(*points_cloud)
        for edge in self.edges:
            # Only edges not belonging to the convex hull may be hidden
            if not (edge.endpoints[0] in cvh and edge.endpoints[1] in cvh):
                m = self._edges3D[edge].midpoint()
                pm = project(m, matrix)  # projected midpoint
                # Check if the midpoint of the tested edge is behind (i.e.
                # deeper) than a face while being inside it
                for f in object3D.faces:
                    pface = [project(v, matrix) for v in f.vertices]
                    if (all([v.z <= m.z for v in f.vertices])
                        and pm not in convex_hull(pm, *pface)
                        and not any(m.belongs_to(s) for s in f.sides)):
                        edge.dashpattern = \
                            config.oblique_projection.DASHPATTERN

        # Setup the vertices' labels
        for vertex in self.vertices:
            edges = sorted(vertices_connexions[vertex],
                           key=lambda edge: Vector(edge).slope360)
            couples = [(edges[n], edges[(n + 1) % len(edges)])
                       for n, _ in enumerate(edges)]
            widest = max(couples,
                         key=lambda couple:
                         Vector(couple[0]).angle_measure(Vector(couple[1])))
            u = Vector(vertex, widest[0].endpoints[1])
            v = Vector(vertex, widest[1].endpoints[1])
            vertex.label_position = \
                tikz_approx_position(u.bisector(v).slope360)

    @property
    def edges(self):
        return self._edges

    @property
    def vertices(self):
        return self._vertices

    @property
    def object3D_name(self):
        return self._object3D_name

    @property
    def draw_vertices(self):
        return self._draw_vertices

    @draw_vertices.setter
    def draw_vertices(self, value):
        if isinstance(value, bool):
            self._draw_vertices = value
        else:
            raise TypeError('draw_vertices must be a boolean; '
                            'found {} instead.'.format(type(value)))

    @property
    def label_vertices(self):
        return self._label_vertices

    @label_vertices.setter
    def label_vertices(self, value):
        if isinstance(value, bool):
            self._label_vertices = value
        else:
            raise TypeError('label_vertices must be a boolean; '
                            'found {} instead.'.format(type(value)))

    def tikz_declarations(self):
        """Return the Points declarations."""
        return '\n'.join([v.tikz_declarations() for v in self.vertices])

    def _tikz_draw_options(self):
        """Unused, as the picture consists of several draw commands."""

    def _tikz_draw_vertices(self):
        return '\n'.join([v.tikz_draw()[0] for v in self.vertices]) + '\n'

    def tikz_drawing_comment(self):
        """Return the comment preceding the Oblique Projection's drawing."""
        output = []
        if self.draw_vertices:
            output.append('% Draw Vertices')
        output.append('% Draw Oblique Projection of {}'
                      .format(self.object3D_name))
        return output

    def tikz_draw(self):
        """
        Return the commands to actually draw the projected 3D object.

        :rtype: list
        """
        output = []
        if self.draw_vertices:
            output.append(self._tikz_draw_vertices())
        drawn_edges = ''
        for edge in self.edges:
            drawn_edges += edge.tikz_draw()[-1] + '\n'
        output.append(drawn_edges)
        return output

    def tikz_label(self):
        """Return the command to write the object's label."""

    def tikz_points_labels(self):
        """Return the command to write the Vertices' labels."""
        if self.label_vertices:
            return '\n'.join([v.tikz_label() for v in self.vertices])
        return ''



# -*- coding: utf-8 -*-

# Mathmaker Lib offers lualatex-printable mathematical objects.
# Copyright 2006-2019 Nicolas Hainaux <nh.techn@gmail.com>

# This file is part of Mathmaker Lib.

# Mathmaker Lib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.

# Mathmaker Lib is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Mathmaker Lib; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""Fractions."""

from math import gcd

from mathmakerlib.calculus.number import Sign, Number
from mathmakerlib.calculus.tools import is_number, is_integer
from mathmakerlib.calculus.tools import prime_factors
from mathmakerlib.core.signed import Signed
from mathmakerlib.core.printable import Printable
from mathmakerlib.core.evaluable import Evaluable
from mathmakerlib.exceptions import StopFractionReduction


class Fraction(Signed, Printable, Evaluable):
    """Fractions."""

    # To keep Fraction immutable, use __new__ not __init__
    def __new__(cls, sign=None, numerator=None, denominator=None,
                from_decimal=None):
        """
        Fraction(5, 8) is equivalent to Fraction('+', 5, 8).
        """
        if from_decimal is not None:
            f = Number(10) ** max(1, from_decimal.fracdigits_nb())
            sign = Sign(from_decimal)
            numerator = (f * from_decimal).standardized()
            denominator = f.standardized()
        elif sign in ['+', '-']:
            if not (is_number(numerator) and is_number(denominator)):
                raise TypeError('Numerator and denominator must be numbers. '
                                'Got {} and {} instead.'
                                .format(type(numerator), type(denominator)))
            if not(is_integer(numerator) and is_integer(denominator)):
                raise TypeError('Numerator and denominator must be integers. '
                                'Got {} and {} instead.'
                                .format(numerator, denominator))
        else:
            if not (is_number(sign) and is_number(numerator)):
                raise TypeError('Numerator and denominator must be numbers. '
                                'Got {} and {} instead.'
                                .format(type(sign), type(numerator)))
            elif not (is_integer(sign) and is_integer(numerator)):
                raise TypeError('Numerator and denominator must be integers. '
                                'Got {} and {} instead.'
                                .format(sign, numerator))
            denominator = numerator
            numerator = sign
            sign = '+'
        # some initialization
        self = object.__new__(cls)
        self._sign = Sign(sign)
        self._numerator = Number(numerator)
        self._denominator = Number(denominator)
        return self

    def __hash__(self):
        return hash(str(self.sign) + str(self.numerator)
                    + str(self.denominator))

    def __eq__(self, other):
        if not isinstance(other, Fraction):
            return False
        return (self.sign == other.sign and self.numerator == other.numerator
                and self.denominator == other.denominator)

    def __ne__(self, other):
        if not isinstance(other, Fraction):
            return True
        return (self.sign != other.sign or self.numerator != other.numerator
                or self.denominator != other.denominator)

    def __lt__(self, other):
        return self.evaluate().__lt__(other)

    def __gt__(self, other):
        return self.evaluate().__gt__(other)

    def __le__(self, other):
        return self.evaluate().__le__(other)

    def __ge__(self, other):
        return self.evaluate().__ge__(other)

    def __repr__(self):
        if self.sign == '+':
            return 'Fraction({}, {})'.format(self.numerator, self.denominator)
        return 'Fraction(\'{}\', {}, {})'.format(self.sign, self.numerator,
                                                 self.denominator)

    @property
    def sign(self):
        return self._sign

    @property
    def numerator(self):
        """Numerator of the Fraction."""
        return self._numerator

    @property
    def denominator(self):
        """Denominator of the Fraction."""
        return self._denominator

    def imprint(self, start_expr=True, variant='latex'):
        s = self.sign
        if s == '+' and start_expr:
            s = ''
        if variant == 'latex':
            return r'{}\dfrac{}{}' \
                .format(s, '{' + str(self.numerator) + '}',
                        '{' + str(self.denominator) + '}')
        elif variant == 'user_input':
            return '{}{}/{}'.format(s, self.numerator, self.denominator)

    def evaluate(self, **kwargs):
        return self.sign.evaluate() \
            * self.numerator.evaluate() / self.denominator.evaluate()

    def is_reducible(self):
        """True if the Fraction is reducible."""
        return gcd(int(self.numerator), int(self.denominator)) > 1

    def reduced(self):
        """Completely reduced Fraction (may return an integer number)."""
        g = gcd(int(self.numerator), int(self.denominator))
        return Fraction(self.sign, self.numerator / g, self.denominator / g)

    def reduced_by(self, n):
        """Return Fraction reduced by n (possibly return a Number)."""
        if not is_number(n):
            raise TypeError('A Fraction can be reduced only by an integer, '
                            'got {} instead.'.format(type(n)))
        if not is_integer(n):
            raise TypeError('A Fraction can be reduced only by an integer, '
                            'got {} instead.'.format(n))
        if not is_integer(self.numerator / n):
            raise ValueError('Cannot divide {} by {} and get an integer.'
                             .format(self.numerator, n))
        if not is_integer(self.denominator / n):
            raise ValueError('Cannot divide {} by {} and get an integer.'
                             .format(self.denominator, n))
        n = abs(n)
        if self.denominator / n == 1:
            return self.sign * self.numerator / n
        if self.denominator / n == -1:
            return Sign('-') * self.sign * self.numerator / n
        return Fraction(self.sign, self.numerator / n, self.denominator / n)

    def reduce(self):
        """
        Return the minimally reduced next Fraction (or Number).

        An exception will be raised if the Fraction is not reducible.
        """
        if not self.is_reducible():
            raise StopFractionReduction(self)
        # lowest prime common divisor
        lpcd = prime_factors(gcd(int(self.numerator),
                                 int(self.denominator)))[0]
        if self.denominator / lpcd == 1:
            return self.sign * self.numerator / lpcd
        return Fraction(self.sign,
                        self.numerator / lpcd, self.denominator / lpcd)
